""" Common stuff for SSO login tests """
# vim: shiftwidth=4 tabstop=4 expandtab

import argparse
import atexit
import getpass
import logging
import os
import os.path
import sys
import traceback
from configparser import ConfigParser

from selenium import webdriver
from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
    TimeoutException,
)
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import WebDriverWait

log = logging.getLogger(__name__)

status_to_exit_code = {
    "OK": 0,
    "WARNING": 1,
    "CRITICAL": 2,
    "UNKNOWN": 3,
}


class EEWebTest:
    """Easter-eggs Web test class"""

    default_timeout = 20
    default_window_size = "1920,1080"
    default_user_data_dir = f"/tmp/.eewebtests.{getpass.getuser()}"  # nosec
    _driver = None
    _config = None
    _options = None
    chrome_options = Options()

    def __init__(self, *args, **kwargs):
        self.parser = argparse.ArgumentParser(*args, **kwargs)

        self.parser.add_argument("-c", "--config-path", help="Configuration file path")
        self.parser.add_argument("-d", "--debug", action="store_true", help="Enable debug mode")
        self.parser.add_argument("-v", "--verbose", action="store_true", help="Enable verbose mode")
        self.parser.add_argument(
            "-D", "--screen-debug", action="store_true", help="Show enbeded Chromium browser window"
        )

        self.sso_opts = self.parser.add_argument_group("SSO options")
        self.sso_opts.add_argument("-u", "--user", dest="sso_user", help="User's login")
        self.sso_opts.add_argument("-p", "--password", dest="sso_passord", help="User's password")
        self.sso_opts.add_argument("-e", "--email", dest="sso_email", help="User's email")
        self.sso_opts.add_argument("-n", "--name", dest="sso_name", help="User's full name")
        self.sso_opts.add_argument("--sso-url", help="Authentic2 SSO URL")

        self.browser_opts = self.parser.add_argument_group("Embedded Chromium browser options")
        self.browser_opts.add_argument(
            "-t",
            "--timeout",
            dest="browser_timeout",
            type=int,
            help=(
                "Timeout delay on interactive delay with the embedded browser "
                f"(in second, default: {self.default_timeout}s)"
            ),
        )
        self.browser_opts.add_argument(
            "-S",
            "--window-size",
            dest="browser_window_size",
            help=f"Window size (default: {self.default_window_size})",
        )
        self.browser_opts.add_argument(
            "--use-shm",
            dest="browser_use_shm",
            action="store_true",
            help="Use shared memory (SHM)",
        )
        self.browser_opts.add_argument(
            "--user-data-dir",
            dest="browser_user_data_dir",
            help=f"User data directory path (default: {self.default_user_data_dir})",
            default=self.default_user_data_dir,
        )

    def init_logging(self):
        """Initialize logging"""
        logging.basicConfig(
            level=(
                logging.DEBUG
                if self.options.debug
                else logging.INFO
                if self.options.verbose
                else logging.WARNING
            ),
            format="%(asctime)s - %(name)s:%(lineno)d - %(levelname)s - %(message)s",
        )

    @property
    def options(self):
        """Parse options"""
        if self._options:
            return self._options

        self._options = self.parser.parse_args()

        self.init_logging()

        if self._options.sso_user and not self._options.sso_password:
            self._options.sso_password = getpass.getpass(
                f"Please enter {self._options.sso_user} password: "
            )

        return self._options

    @property
    def config(self):
        """Parse config file"""
        if self._config:
            return self._config
        self._config = ConfigParser(defaults={"hash": "#"})
        if not self.options.config_path:
            default_config_path = os.path.join(
                os.path.dirname(os.path.realpath(sys.argv[0])), "config.ini"
            )
            if os.path.exists(default_config_path):
                log.info("Default configuration file found (%s), use it", default_config_path)
                self.options.config_path = default_config_path
            else:
                log.warning("Default configuration file not found (%s)", default_config_path)
        if self.options.config_path:
            try:
                log.info("Load configuration file %s", self.options.config_path)
                self._config.read(self.options.config_path, encoding="utf8")
            except Exception:  # pylint: disable=broad-except
                self.clean_exit(
                    f"Fail to load configuration file ({self.options.config_path})", status="OK"
                )
        return self._config

    def get_option(self, section, option, default=None):
        """Get one option value"""
        opt = f"{section}_{option}" if section else option
        value = None
        if hasattr(self.options, opt):
            value = getattr(self.options, opt)
            log.debug("get_option(%s, %s): from options %s='%s'", section, option, opt, value)

        if section and (
            value is None or (hasattr(self.options, opt) and value == self.parser.get_default(opt))
        ):
            if self.config.has_section(section) and self.config.has_option(section, option):
                value = self.config.get(section, option)
                log.debug("get_option(%s, %s): from config='%s'", section, option, value)
            else:
                log.debug("get_option(%s, %s): no section %s in config", section, option, section)

        if value is None:
            value = default
            log.debug("get_option(%s, %s): use default value (%s)", section, option, default)
        log.debug("get_option(%s, %s): result='%s'", section, option, value)
        return value

    @property
    def driver(self):
        """Initialize driver"""
        if self._driver:
            return self._driver
        log.info("Initializing driver")

        # Disable sandbox when running as root to avoid Chromium crash on starting
        # Source: https://stackoverflow.com/questions/50642308/webdriverexception-unknown-error-
        # devtoolsactiveport-file-doesnt-exist-while-t
        if os.geteuid() == 0:
            self.chrome_options.add_argument("--no-sandbox")

        # Disable SHM usage to fix crash problem running in Docker
        # Source: https://stackoverflow.com/questions/53902507/unknown-error-session-deleted-
        # because-of-page-crash-from-unknown-error-cannot
        if not self.get_option("browser", "use_shm"):
            self.chrome_options.add_argument("--disable-dev-shm-usage")
            log.info("Disable SHM usage")
        else:
            log.info("Use SHM")

        # Enable headless mode unless screen debugging is enabled
        if not self.options.screen_debug:
            self.chrome_options.add_argument("--headless")
            self.chrome_options.add_argument(
                "--window-size="
                + self.get_option("browser", "window_size", default=self.default_window_size)
            )

        # Set custom user data directory path
        user_data_dir = self.get_option("browser", "user_data_dir")
        if user_data_dir:
            if not os.path.exists(user_data_dir):
                log.debug("Create user data directory (%s)", user_data_dir)
                os.mkdir(user_data_dir, mode=0o0700)
            self.chrome_options.add_argument(f"--user-data-dir={user_data_dir}")
            # Also set TMPDIR environment variable to avoid leaving stuff in /tmp on crash
            # See: https://github.com/FuckTheWorld/chromedriver/issues/721
            os.environ["TMPDIR"] = os.path.join(user_data_dir, "tmp")
            if not os.path.exists(os.environ["TMPDIR"]):
                log.debug("Create temporary directory (%s)", os.environ["TMPDIR"])
                os.mkdir(os.environ["TMPDIR"], mode=0o0700)

        # Start Chrome webdriver
        log.debug("Start enbeded Web browser with parameters %s", self.chrome_options)
        self._driver = webdriver.Chrome(options=self.chrome_options)
        self._driver.maximize_window()

        # Close chromium at exist
        atexit.register(self._driver.quit)
        return self._driver

    def sso_login(self):
        """SSO login"""
        # Check configuration
        for opt in ("url", "user", "password"):
            if not self.get_option("sso", opt):
                self.clean_exit(f"SSO {opt} not provided", status="UNKNOWN")

        # Get login page
        log.info("Get SSO login page (%s)", self.sso_url)
        self.driver.get(self.sso_url)

        # Fill the form
        # pylint: disable=duplicate-code
        self.locate(
            "username",
            by=By.NAME,
            value=self.get_option("sso", "user"),
            label="username input on SSO login page",
            submit=False,
        )
        self.locate(
            "password",
            by=By.NAME,
            value=self.get_option("sso", "password"),
            label="password input on SSO login page",
            submit=True,
        )

        # Check Your account button
        self.locate("a[href='/accounts/']", label="account button on SSO homepage")

    def clean_exit(self, msg, extra_lines=None, exc=None, status=None):
        """Error handler for clean exit"""
        status = (status or "WARNING").upper()
        print(f"{status} - {msg}")
        if status == "UNKNOWN":
            log.debug(self.driver.execute_script("return document.body.innerHTML;"))
        if exc:
            traceback.print_exception(exc)
        print("\n".join(extra_lines or []))
        if self.options.screen_debug:
            input("Press [enter] to stop the web browser and exit")

        sys.exit(status_to_exit_code.get(status, 3))

    @staticmethod
    def _all_off_elements(check_one, elements):
        """
        Check all of specified elements respect the condition implemented by the provided check_one
        condition
        """
        results = []
        for element in elements:
            result = check_one(element)
            if result is False:
                return False
            results.append(result)
        return results

    def _one_or_multiple(self, locator, multiple, check_one):
        """Helper to find and check one or multiple elements"""

        def condition(driver):
            return (
                self._all_off_elements(check_one, driver.find_elements(*locator))
                if multiple
                else check_one(driver.find_element(*locator))
            )

        return condition

    def _non_empty_text(self, locator, multiple=False):  # pylint: disable=too-few-public-methods
        """Helper to use with WebDriverWait to retrieve non-empty element(s)"""

        def check_one(element):
            try:
                if element.tag_name == "input":
                    return element if element.get_attribute("value") else False
                return element if element.text else False
            except StaleElementReferenceException:
                return False

        return self._one_or_multiple(locator, multiple, check_one)

    def _visible_element(self, locator, multiple=False):  # pylint: disable=too-few-public-methods
        """Helper to use with WebDriverWait to retrieve visible element(s)"""

        def check_one(element):
            try:
                return element if element.is_displayed() else False
            except StaleElementReferenceException:
                return False

        return self._one_or_multiple(locator, multiple, check_one)

    @staticmethod
    def _present_element(locator, multiple=False):  # pylint: disable=too-few-public-methods
        """Helper to use with WebDriverWait to retrieve visible element(s)"""

        def condition(driver):
            return driver.find_elements(*locator) if multiple else driver.find_element(*locator)

        return condition

    def _click_on_element(self, locator, multiple=False):  # pylint: disable=too-few-public-methods
        """Helper to use with WebDriverWait to retrieve an non-empty element"""

        def check_one(element):
            try:
                element.click()
                return True
            except StaleElementReferenceException:
                return False

        return self._one_or_multiple(locator, multiple, check_one)

    def locate(
        self,
        pattern,
        by=None,
        label=None,
        non_empty=False,
        value=None,
        submit=None,
        click=None,
        allow_missing=False,
        wait_delay=None,
        multiple=False,
    ):
        """Locate an element"""
        by = by or By.CSS_SELECTOR
        label = label or pattern
        wait_delay = wait_delay or self.get_option(
            "browser", "timeout", default=self.default_timeout
        )
        try:
            log.info("Locate %s", label)
            if non_empty:
                wait_condition = self._non_empty_text((by, pattern), multiple=multiple)
            elif value or click:
                wait_condition = self._visible_element((by, pattern), multiple=multiple)
            else:
                wait_condition = self._present_element((by, pattern), multiple=multiple)
            try:
                element = WebDriverWait(self.driver, wait_delay).until(wait_condition)
            except (NoSuchElementException, TimeoutException) as err:
                if allow_missing:
                    log.info("%s is missing, ignore it", label)
                    return None
                self.clean_exit(f"Fail to locate {label} ({err})", status="UNKNOWN")

            # Enter value in input element
            if value:
                log.info("Set %s value", label)
                try:
                    element.clear()
                    element.send_keys(value)
                except Exception as exc:  # pylint: disable=broad-exception-caught
                    self.clean_exit(f"Exception occurred filling {label}", exc, status="UNKNOWN")

                if submit is not False:
                    log.info("Submit form")
                    try:
                        element.send_keys(Keys.RETURN)
                    except Exception as exc:  # pylint: disable=broad-exception-caught
                        self.clean_exit(
                            f"Exception occurred submiting {label} form",
                            exc,
                            status="UNKNOWN",
                        )

            if click:
                log.info("Click on %s", label)
                try:
                    WebDriverWait(self.driver, wait_delay).until(
                        self._click_on_element((by, pattern))
                    )
                except Exception as exc:  # pylint: disable=broad-exception-caught
                    self.clean_exit(
                        f"Exception occurred clicking on {label}", exc, status="UNKNOWN"
                    )

            return element
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.clean_exit(f"Exception occurred locating {label}", exc, status="UNKNOWN")
        return None

    def __getattr__(self, attr):
        if "_" in attr:
            parts = attr.split("_", maxsplit=1)
            value = self.get_option(parts[0], parts[1], default="__NOTSET__")
            if value != "__NOTSET__":
                return value
        if self._driver and hasattr(self._driver, attr):
            return getattr(self._driver, attr)
        raise AttributeError(attr)
